#include "jsm.h"

/* mod_offline_count must go before mod_presence 
  use mod_offline or mod_offline_count */

#define NS_OFFLINE_COUNT "jabber:x:offline-count"

#define OFFLINE_ERASE_OTHER_EVENTS 1

/* handle an offline message */
mreturn mod_offline_count_message(mapi m)
{
#ifdef OFFLINE_SESSION_CHECK
    session top;
#endif
    xmlnode cur = NULL, cur2;
    char str[10];
    char *id;

	
    /* if there's an existing session, just give it to them */
#ifdef OFFLINE_SESSION_CHECK
    if ((top = js_session_primary_sem(m->user)) != NULL) {
	  js_session_to(top, m->packet);
	  return M_HANDLED;
    }
#endif

    switch(jpacket_subtype(m->packet))
    {
    case JPACKET__NONE:
    case JPACKET__ERROR:
    case JPACKET__CHAT:
        break;
    default:
        return M_PASS;
    }

	/* look for event messages */
	id = xmlnode_get_attrib(m->packet->x,"id");
	if (id) {
	  for(cur = xmlnode_get_firstchild(m->packet->x); cur != NULL; cur = xmlnode_get_nextsibling(cur))
		if(NSCHECK(cur,NS_EVENT))  {
		  if(xmlnode_get_tag(cur,"id") != NULL)
			return M_PASS; /* bah, we don't want to store events offline (XXX: do we?) */
		  if(xmlnode_get_tag(cur,"offline") != NULL)
			break; /* cur remaining set is the flag */
		}	  
	  /* erase any other events */
#ifdef OFFLINE_ERASE_OTHER_EVENTS
	  if (cur)
		xmlnode_hide(cur);
#endif	  
	}

	
    log_debug("mod_offline","handling message for %s", m->user->user);
	
    if((cur2 = xmlnode_get_tag(m->packet->x,"x?xmlns=" NS_EXPIRE)) != NULL) {

	  /* do not put to offline message with expire < 2 sec */
	  if(j_atoi(xmlnode_get_attrib(cur2, "seconds"),0) < 2)
		  return M_PASS;
	  
	  sprintf(str,"%d",(int)time(NULL));
	  xmlnode_put_attrib(cur2,"stored",str);
    }
    jutil_delay(m->packet->x,"Offline Storage");
    

    /* try to put */
    if(xdb_act(m->si->xc, m->user->id, NS_OFFLINE, "insert", NULL, m->packet->x)) {      
      /* free messages from us */
      if(jid_cmpx(m->packet->from, m->user->id, JID_USER | JID_SERVER) == 0) {
	log_alert("offline_error","Offline message from own user %s generated offline storage error", m->user->user);
	xmlnode_free(m->packet->x);
	return M_HANDLED;
      }
      
      /* replay message */
      jutil_tofrom(m->packet->x);
      
      /* hide everythink */
      for(cur = xmlnode_get_firstchild(m->packet->x); cur != NULL; cur = xmlnode_get_nextsibling(cur))
	xmlnode_hide(cur);
      
      /* build message */
      {
	char *body;
#ifdef FORWP
	body = "System: Wiadomosci nie dostarczono. Skrzynka wiadomosci jest pelna.";
#else
	body = "System: Message wasn't delivered. Offline storage size was exceeded.";
#endif
	xmlnode_insert_cdata (xmlnode_insert_tag (m->packet->x, "body"), body, -1);
      }
      log_alert("offline_error","Offline message not delivered from user %s",jid_full(m->packet->from));
      
      js_deliver(m->si, jpacket_reset(m->packet));	  
      
      return M_HANDLED;	  
    }

    /* count ++ */
    cur2 = xdb_get(m->si->xc, m->user->id, NS_OFFLINE_COUNT);
    if (cur2) {
      char buf[10];
      int count;
      char *jid_from;
      int change = 1;

      jid_from = jid_full(jid_user(m->packet->from));

      count = j_atoi(xmlnode_get_attrib(cur2,"count"), 0);
      if (count > 0) {
	if (j_strcmp( jid_from, xmlnode_get_attrib(cur2,"lastfrom"))==0)
	  change = 0;
      }
      
      if (change) {
	count++;
	snprintf(buf,9,"%d",count);

	xmlnode_put_attrib(cur2, "count", buf);
	xmlnode_put_attrib(cur2, "lastfrom", jid_from);
	xdb_set(m->si->xc, m->user->id, NS_OFFLINE_COUNT, cur2);
      }
      xmlnode_free(cur2);
    }
    else {
      cur2 = xmlnode_new_tag("messagecount");
      xmlnode_put_attrib(cur2, "count", "1");
      xmlnode_put_attrib(cur2, "lastfrom", jid_full(jid_user(m->packet->from)));
      xdb_set(m->si->xc, m->user->id, NS_OFFLINE_COUNT, cur2);
      xmlnode_free(cur2);
    }
    
    if(cur != NULL)  {
      log_debug(ZONE,"offline event");
      
      cur = util_offline_event(m->packet->x,
			       jid_full(m->user->id),
			       xmlnode_get_attrib(m->packet->x,"from"),
			       id);	  
      js_deliver(m->si, jpacket_new(cur));	  
    }
    else {
      xmlnode_free(m->packet->x);
    }
    
    return M_HANDLED;
}

/* just breaks out to our message/presence offline handlers */
mreturn mod_offline_count_handler(mapi m, void *arg)
{
    if(m->packet->type == JPACKET_MESSAGE) return mod_offline_count_message(m);

    return M_IGNORE;
}

/* watches for when the user is available and sends out offline messages */
void mod_offline_count_out_available(mapi m)
{
    xmlnode opts, cur, x;
    int now = time(NULL);
    int expire, stored, diff;
    char str[10];

    log_debug("mod_offline","avability established, check for messages");

    if((opts = xdb_get(m->si->xc, m->user->id, NS_OFFLINE)) == NULL)
        return;

    /* check for msgs */
    for(cur = xmlnode_get_firstchild(opts); cur != NULL; cur = xmlnode_get_nextsibling(cur))
    {
        /* check for expired stuff */
        if((x = xmlnode_get_tag(cur,"x?xmlns=" NS_EXPIRE)) != NULL)
        {
            expire = j_atoi(xmlnode_get_attrib(x,"seconds"),0);
            stored = j_atoi(xmlnode_get_attrib(x,"stored"),now);
            diff = now - stored;
            if(diff >= expire)
            {
                log_debug(ZONE,"dropping expired message %s",xmlnode2str(cur));
                xmlnode_hide(cur);
                continue;
            }
            sprintf(str,"%d",expire - diff);
            xmlnode_put_attrib(x,"seconds",str);
            xmlnode_hide_attrib(x,"stored");
        }
        js_session_to(m->s,jpacket_new(xmlnode_dup(cur)));
        xmlnode_hide(cur);
    }
    /* messages are gone, save the new sun-dried opts container */
    xdb_set(m->si->xc, m->user->id, NS_OFFLINE, NULL);

    /* clear counter */
    xdb_set(m->si->xc, m->user->id, NS_OFFLINE_COUNT, NULL);
    
    xmlnode_free(opts);
}

mreturn mod_offline_count_out(mapi m, void *arg)
{
    if(m->packet->type != JPACKET_PRESENCE) return M_IGNORE;

    if(js_online(m))
        mod_offline_count_out_available(m);

    return M_PASS;
}

/* sets up the per-session listeners */
mreturn mod_offline_count_session(mapi m, void *arg)
{
    log_debug(ZONE,"session init");

    js_mapi_session(es_OUT, m->s, mod_offline_count_out, NULL);

    return M_PASS;
}

JSM_FUNC void mod_offline_count(jsmi si)
{
    log_debug("mod_offline","init");
    js_mapi_register(si, e_OFFLINE, mod_offline_count_handler, NULL);
    js_mapi_register(si, e_SESSION, mod_offline_count_session, NULL);
}
