/* --------------------------------------------------------------------------
   WPPHP

   AUTHORS:
   Lukasz Karwacki <lukasm@wp-sa.pl>
   Wojciech Meller <wmeler@wp-sa.pl>

   -------------------------------------------------------------------------- */

#include "wpphp.h"

/* parse user data */
typedef struct wpphp_parser_struct {
  pool p;
  int root;
  xmlnode current;
  void *parser;
} *wpphp_parse,_wpphp_parse;


void _wpphp_xstream_startElement(wpphp_parse parser, 
							   const char* name, 
							   const char** atts)
{
  if (parser->current == NULL) {
	pool p = pool_heap(5 * 1024); 
	parser->current = xmlnode_new_tag_pool(p, name);
	xmlnode_put_expat_attribs(parser->current, atts);	

	if (parser->root == 0) {
	  parser->root = 1;
	  xmlnode_free(parser->current);
	  parser->current = NULL;
	}
  }
  else   {
	parser->current = xmlnode_insert_tag(parser->current, name);
	xmlnode_put_expat_attribs(parser->current, atts);
  }
}

void _wpphp_xstream_endElement(wpphp_parse parser,
							 const char* name)
{
  xmlnode parent;

  if (parser->current == NULL) {
	log_debug(ZONE,"error");
	return;
  }

  parent = xmlnode_get_parent(parser->current);
  
  if (parent == NULL) {	
	log_debug(ZONE,"got packet %s",xmlnode2str(parser->current));
	
	logger("packet_out","php",xmlnode2str(parser->current));
	
	deliver(dpacket_new(parser->current),NULL);
  }
  
  parser->current = parent;  
}

void _wpphp_xstream_CDATA(wpphp_parse parser, const char* cdata, int len)
{
  if (parser->current != NULL)
	xmlnode_insert_cdata(parser->current, cdata, len);
}

void _wpphp_xstream_cleanup(void* arg)
{
  wpphp_parse parser = (wpphp_parse)arg;

  if (parser->current) {
	xmlnode_free(parser->current);
	parser->current = NULL;
  }
  
  if (parser->parser) {
	XML_ParserFree(parser->parser);
	parser->parser = NULL;
  }
}

void _wpphp_xstream_init(wpphp_parse parser)
{
  if (parser != NULL)   {	
	log_debug(ZONE,"init parser");
	/* Initialize the parser */
	parser->parser = XML_ParserCreate(NULL);
	XML_SetUserData(parser->parser, (void *)parser);
	XML_SetElementHandler(parser->parser, 
						  (void*)_wpphp_xstream_startElement, 
						  (void*)_wpphp_xstream_endElement);
	XML_SetCharacterDataHandler(parser->parser, 
								(void*)_wpphp_xstream_CDATA);
	pool_cleanup(parser->p, _wpphp_xstream_cleanup, (void*)parser);
  }
}

void wp_php_parser(void **parser,const char *str,int length)
{
  wpphp_parse wpparser = *parser;

  log_debug(ZONE,"parser %s len %d",str,length);
  
  if ((str==NULL) || (length <= 0)) {
	log_alert(ZONE,"bad PHP output");
	return;
  }

  log_debug(ZONE,"parser %s len %d",str,length);

  if (wpparser == NULL) {
	pool p = pool_new();
	log_debug(ZONE,"init wpparser struct");
	wpparser = pmalloco(p, sizeof(_wpphp_parse));
	wpparser->p = p;		
	*parser = wpparser;
  }

  if(wpparser->parser == NULL)   {
	_wpphp_xstream_init(wpparser);
	wpparser->root = 0;
  }

  if( XML_Parse(wpparser->parser, str, length, 0) == 0) {
	log_alert("parser_error","packet %s ; error %s",
			  str,
			  (char *)XML_ErrorString(XML_GetErrorCode(wpparser->parser)));
	_wpphp_xstream_cleanup(wpparser);
  }
}

void wp_php_parser_flush(void **parser) 
{
  wpphp_parse wpparser = *parser;

  log_debug(ZONE,"flush");
  
  if (wpparser)
    _wpphp_xstream_cleanup(wpparser);
}




