/*
   +----------------------------------------------------------------------+
   | PHP version 4.0                                                      |
   +----------------------------------------------------------------------+
   | Copyright (c) 1997-2001 The PHP Group                                |
   +----------------------------------------------------------------------+
   | This source file is subject to version 2.02 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
   | available at through the world-wide-web at                           |
   | http://www.php.net/license/2_02.txt.                                 |
   | If you did not receive a copy of the PHP license and are unable to   |
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
   +----------------------------------------------------------------------+
   | Author: Wojciech Meler wmeler@wp-sa.pl                               |
   |         Lukas karwacki lukasm@wp-sa.pl                               |
   +----------------------------------------------------------------------+
*/


#include "php.h"
#include "SAPI.h"
#include "php_main.h"
#include "php_variables.h"
#include "php_jabber.h"
#include "queues.h"

static int sapi_jabber_ub_write(const char *str, uint str_length TSRMLS_DC) {
	wp_php_parser(&JG(parser),str,str_length);
	return str_length;
}

static int sapi_jabber_send_headers(sapi_headers_struct *sapi_headers TSRMLS_DC) {
    return SAPI_HEADER_SENT_SUCCESSFULLY;
}

static int sapi_jabber_read_post(char *buffer, uint count_bytes TSRMLS_DC) {  
  return 0;
}

static char *sapi_jabber_read_cookies(TSRMLS_D) {
  return NULL;
}

static void sapi_jabber_register_variables(zval *track_vars_array TSRMLS_DC) {
  php_packet pp;

  
  pp = JG(packet);

  if (pp) {
	char buf[11];
	if (pp->packet) {
	  php_register_variable("jabber_packet", pp->packet, track_vars_array TSRMLS_CC);
	  snprintf(buf,10,"%d",pp->len);
	  php_register_variable("jabber_packet_len", buf, track_vars_array TSRMLS_CC);
	}

	snprintf(buf,10,"%d",pp->type);
	php_register_variable("jabber_packet_type",buf, track_vars_array TSRMLS_CC);
	snprintf(buf,10,"%d",pp->subtype);
	php_register_variable("jabber_packet_subtype", buf, track_vars_array TSRMLS_CC);
	
	if (pp->to) {
	  php_register_variable("jabber_packet_to", pp->to, track_vars_array TSRMLS_CC);
	}
	if (pp->from) {
	  php_register_variable("jabber_packet_from", pp->from, track_vars_array TSRMLS_CC);
	}
	
	if (pp->arg1) {
	  php_register_variable("jabber_packet_arg1", pp->arg1, track_vars_array TSRMLS_CC);
	}
	
	if (pp->arg2) {
	  php_register_variable("jabber_packet_arg2", pp->arg2, track_vars_array TSRMLS_CC);
	}
  }
  
  php_register_variable("SERVER_SOFTWARE", "JABBER", track_vars_array TSRMLS_CC);
  php_register_variable("PHP_SELF", "", track_vars_array TSRMLS_CC);
}

static void sapi_jabber_log_msg(char * msg) {
    TSRMLS_FETCH();
	log_alert("PHP","%s",msg);
}

static sapi_module_struct jabber_sapi_module = {
	"jabber",
	"Jabber PHP module",
	php_module_startup,
	php_module_shutdown_wrapper,
	NULL,									/* activate */
	NULL,									/* deactivate */
	sapi_jabber_ub_write,
	NULL, /* flush */
	NULL,									/* get uid */
	NULL,									/* getenv */
	php_error,
	NULL,
	sapi_jabber_send_headers,
	NULL,
	sapi_jabber_read_post,
	sapi_jabber_read_cookies,
	sapi_jabber_register_variables,
	sapi_jabber_log_msg,
	NULL,						/* Block interruptions */
	NULL,						/* Unblock interruptions */
	STANDARD_SAPI_MODULE_PROPERTIES
};

static void jabber_module_main(TSRMLS_D) {
    zend_file_handle file_handle;
    zval *local_retval=NULL;
	
    if (php_request_startup(TSRMLS_C) == FAILURE) {
	  return;
    }
	
    PG(expose_php) = 0;
    PG(display_errors)=0;
	
    file_handle.type = ZEND_HANDLE_FILENAME;
    file_handle.filename = JG(packet)->script;
    file_handle.free_filename = 0;
    file_handle.opened_path = NULL;
    zend_try {
	  php_execute_script(&file_handle TSRMLS_CC);
    } zend_end_try();
    php_request_shutdown(0);
}

static void jabber_request_ctor(TSRMLS_D) {

    SG(request_info).path_translated = NULL;
    SG(request_info).request_uri = NULL;
    SG(request_info).request_method = "GET";
    SG(sapi_headers).http_response_code = 200;
    SG(request_info).content_type = pstrdup(JG(packet)->p,"text/xml");
    SG(request_info).content_length = JG(packet)->len;

    SG(request_info).no_headers=1;
}

static void jabber_request_dtor(TSRMLS_D) {
    pool_free(JG(packet)->p);
    wp_php_parser_flush(&JG(parser));
}

queue_s queue;

void jabber_php_queue_packet(php_packet pp)
{
    queue_put(&queue,pp);
}


void *worker_thread(void * arg) {
    int do_work = 1;
    TSRMLS_FETCH();

    SG(server_context)= (void*)1;

    while (do_work) {
	  JG(packet) = queue_get_wait(&queue);
	  if (JG(packet)) {
	    jabber_request_ctor(TSRMLS_C);
	    jabber_module_main(TSRMLS_C);
	    jabber_request_dtor(TSRMLS_C);
	  }
    }
}

void jabber_php_add_thread(void) {
   pthread_t *thread;

   thread = (pthread_t *) malloc(sizeof(pthread_t));
   pthread_create(thread,NULL,worker_thread,NULL);
}

void jabber_php_shutdown(void)
{
	TSRMLS_FETCH();

	jabber_sapi_module.shutdown(&jabber_sapi_module);
	sapi_shutdown();
	tsrm_shutdown();
}

static void php_jabber_init_globals(php_jabber_globals *globals TSRMLS_DC) { 
  globals->packet=NULL;
  globals->parser =NULL;
}

static int jabber_globals_id;

void jabber_php_init(void) {
    init_queue(&queue);
    tsrm_startup(1, 1, 0, NULL);
    ts_allocate_id(&jabber_globals_id, sizeof(php_jabber_globals),(ts_allocate_ctor) php_jabber_init_globals,NULL);
    sapi_startup(&jabber_sapi_module);
    php_module_startup(&jabber_sapi_module,NULL,0);
    jabber_php_add_thread();
}

